/*
 *
 * Copyright (C) 2001, 2002 David Leuschner, Stefan Heimann
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307
 * USA
 *
 */

package org.cantaloop.cgimlet.handler;

import java.io.File;
import java.io.IOException;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.cantaloop.cgimlet.Assertions;
import org.cantaloop.cgimlet.CodeGenerationException;
import org.cantaloop.cgimlet.ComponentHandler;
import org.cantaloop.cgimlet.OptionReader;
import org.cantaloop.cgimlet.OptionReaderConfigurator;
import org.cantaloop.cgimlet.Options;
import org.cantaloop.cgimlet.ProjectGenerator;
import org.cantaloop.cgimlet.PropertyParser;
import org.cantaloop.cgimlet.XMLUtils;

import org.cantaloop.cgimlet.lang.CGUtils;
import org.cantaloop.cgimlet.lang.ClassTemplate;
import org.cantaloop.cgimlet.lang.CodeWriter;
import org.cantaloop.cgimlet.lang.ConstructorTemplate;
import org.cantaloop.cgimlet.lang.FieldTemplate;
import org.cantaloop.cgimlet.lang.LanguageFactory;
import org.cantaloop.cgimlet.lang.LanguageFactory;
import org.cantaloop.cgimlet.lang.MethodTemplate;
import org.cantaloop.cgimlet.lang.Type;

import org.cantaloop.cgimlet.lang.java.JLanguageFactory;
import org.cantaloop.cgimlet.lang.java.JLanguageFactory;

import org.cantaloop.tools.logging.Logger;

import org.dom4j.Document;
import org.dom4j.Element;
import org.dom4j.Namespace;
import org.dom4j.QName;

/**
 * Handler that generates a java bean. 
 *
 *
 * @created 2002-01-18 10:09:55 CET
 *
 * @author <a href="mailto:david@cantaloop.org">David Leuschner</a>
 * @author <a href="mailto:stefan@cantaloop.org">Stefan Heimann</a>
 *
 * @version @version@ ($Revision: 1.10 $)
 */

public class BeanHandler implements ComponentHandler {

  public static final String NAMESPACE_URI =
    "http://www.cantaloop.org/xmlns/cgimlet/beangen";
  public static final Namespace NAMESPACE = new Namespace(null, NAMESPACE_URI);
  
  private ProjectGenerator m_gen;
  private Options m_opt;
  private Logger m_log;
  private LanguageFactory m_lang;
  private CGUtils m_utils;
  
  public BeanHandler(){
    m_opt = new Options();
    m_lang = JLanguageFactory.getInstance();
    m_utils = m_lang.getUtils();
  }

  public void init(ProjectGenerator gen) {
    m_gen = gen;
    m_log = gen.getLogger(this);
    m_log.info("Handler initialized.");
  }

  public void configure(Element elem) {
    OptionReaderConfigurator config = new OptionReaderConfigurator(NAMESPACE);
    OptionReader reader = new OptionReader(config, elem, m_opt);
    reader.read();
    m_log.info("Handler configured.");
  }

  public Object generate(Element elem) {
    String href = elem.attributeValue("href");
    Assertions.assertNotNull(href,"No `href' attribute found " +
                             "in the content section " +
                             "of the project file.");
    File f = m_gen.resolveFilename(href);
    Document d = XMLUtils.parse(f);
    Element root = d.getRootElement();
    PropertyParser parser = new PropertyParser(NAMESPACE);
    parser.apply(root);
    m_log.info("Read and parsed " + href + ".");

    List results = new LinkedList();
    for(Iterator it=root.elementIterator(new QName("java-bean", NAMESPACE));
        it.hasNext(); ) {
      Element bean = (Element)it.next();
      results.add(generateBean(bean));
    }
    return results;
  }

  private File generateBean(Element bean) {
    String name = bean.attributeValue("name");
    String pkg =  bean.attributeValue("package");    
    Assertions.assertNotNull(name, "Every bean needs a name.");

    ClassTemplate ct = m_lang.getClassTemplate(pkg, name);
    m_log.info("Generating bean " + ct.getFullyQualifiedName());
    /* Every bean needs a no-arg constructor. */
    ct.addConstructor();
    m_log.debug("Added no-arg constructor.");

    for(Iterator it=bean.elementIterator(new QName("import", NAMESPACE));
        it.hasNext(); ) {
      Element imp = (Element)it.next();
      String impStr = imp.getTextTrim();
      ct.addImport(impStr);
      m_log.debug("Added import `" + impStr + "'.");
    }
    
    OptionReaderConfigurator config =
      new OptionReaderConfigurator(new QName("field", NAMESPACE),
                                   "name", "type");
    OptionReader reader = new OptionReader(config, bean);
    reader.read();
    
    Options fields = reader.getOptions();
    for(Iterator it=fields.entrySet().iterator(); it.hasNext(); ) {
      Map.Entry entry = (Map.Entry)it.next();
      String fName = (String)entry.getKey();
      /* FIXME: not language independent */
      Type fType = new Type((String)entry.getValue()); 
      FieldTemplate field = m_lang.getFieldTemplate(fType, "m_" + fName);
      ct.addField(field);
      ct.addMethod(m_utils.getter(fName, field));
      ct.addMethod(m_utils.setter(fName, field));
      m_log.debug("Added field + `" + fName + "' of type `" + fType + "'.");
    }

    Element constructors = bean.element("constructors");
    if(constructors != null) {
      for(Iterator it = constructors.elementIterator(new QName("constructor",
                                                               NAMESPACE));
          it.hasNext(); ) {
        Element constrElem = (Element)it.next();
        ConstructorTemplate constr = ct.addConstructor();
        for(Iterator innerIt=constrElem.elementIterator(new QName("arg",
                                                                  NAMESPACE));
            innerIt.hasNext(); ) {
          String cName = ((Element)innerIt.next()).getTextTrim();
          /* FIXME: not language independent */
          Type cType = new Type(fields.getOption(cName));
          Assertions.assertNotNull(cType, "No field with name `" + cName +
                                   "' found. " + "Cannot set an unknown " +
                                   "field in a constructor.");
          constr.addParameter(cType, cName);
          constr.append("m_" + cName + " = " + cName + ";");
        }
        m_log.debug("Added constructor with paramters " +
                    constr.parameterToString() + ".");
      }
    }
     
    String outputDir = m_opt.getOption("output", ".");
    m_log.debug("Output directory is `" + outputDir + "'.");
    CodeWriter w = m_gen.getCodeWriter(outputDir, m_lang);
    try {
      File dest = (File)w.write(ct);
      m_log.info("Class " + ct.getName() + " written to "
                 + dest.getCanonicalPath());
      return dest;
    } catch(IOException e) {
      throw new CodeGenerationException("Error while writing class `"
                                        + ct.getFullyQualifiedName() + ".", e);
    }
  }
}
